library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.emp_data_types.all;

use work.ipbus.all;
use work.ipbus_reg_types.all;
use work.mprocessor_data_types.all;

use work.ipbus_decode_link_aggregator.all;


entity LinkAggregatorIPBus is
    generic (
        N_LINKS : integer;
        INDEX   : integer := 0
    );
    port (
        --- Input Ports ---
        clk_p              : in std_logic;
        packet_start       : in std_logic;
        links_in           : in ldata(N_LINKS - 1 downto 0);
        --- Output Ports ---
        links_out          : out ldata(7 downto 0) := (others => LWORD_NULL);
        uncompressed_stubs : out ldata(N_LINKS * 8 - 1 downto 0) := (others => LWORD_NULL);
        readout_reset      : out std_logic := '0';
        tracking_stubs     : out ldata(5 downto 0) := (others => LWORD_NULL);
        --- IPBus Ports ---
        ipb_clk            : in  std_logic;
        ipb_rst            : in  std_logic;
        ipb_in             : in  ipb_wbus;
        ipb_out            : out ipb_rbus
    );
end LinkAggregatorIPBus;


architecture Behavorial of LinkAggregatorIPBus is

    signal ipb_to_slaves           : ipb_wbus_array(N_SLAVES - 1 downto 0);
    signal ipb_from_slaves         : ipb_rbus_array(N_SLAVES - 1 downto 0);
    signal channel_sel             : ipb_reg_v(0 downto 0);
    signal ipb_chain               : ipbdc_bus_array(8 downto 0);

    -- signal link_aggregator_status_registers : ipb_reg_v(4 - 1 downto 0) := (others => (others => '0'));
    signal control_registers                : ipb_reg_v(2 downto 0)             := (others => (others => '0'));
    signal mprocessor_monitoring_registers  : ipb_reg_v(8*N_LINKS - 1 downto 0) := (others => (others => '0'));
    signal rate_monitoring_registers        : ipb_reg_v(2 - 1 downto 0)         := (others => (others => '0'));
    signal bx_valid_count                   : ipb_reg_v(1 - 1 downto 0)         := (others => (others => '0'));

    signal masked_links                     : ldata(N_LINKS - 1 downto 0)       := (others => LWORD_NULL);
    signal thresholds_l                     : tIntegerArray                     := (others => 0);
    signal thresholds_u                     : tIntegerArray                     := (others => 0);

    signal tracking_event_count : std_logic_vector(31 downto 0) := (others => '0');

begin

    --==============================--
    -- IPBus
    --==============================--

    --==============================--
    fabric: entity work.ipbus_fabric_sel
    --==============================--
    generic map(
        NSLV            => N_SLAVES,
        SEL_WIDTH       => IPBUS_SEL_WIDTH
    )
    port map(
        ipb_in          => ipb_in,
        ipb_out         => ipb_out,
        sel             => ipbus_sel_link_aggregator(ipb_in.ipb_addr),
        ipb_to_slaves   => ipb_to_slaves,
        ipb_from_slaves => ipb_from_slaves
    );

    --==============================--
    channel_ctrl: entity work.ipbus_reg_v
    --==============================--
    generic map(
        N_REG           => 1
    )
    port map(
        clk             => ipb_clk,
        reset           => ipb_rst,
        ipbus_in        => ipb_to_slaves(N_SLV_CHAN_SEL),
        ipbus_out       => ipb_from_slaves(N_SLV_CHAN_SEL),
        q               => channel_sel,
        qmask           => (0 => X"0000007f")
    );

    --==============================--
    ControlInstance: entity work.ipbus_ctrlreg_v
    --==============================--
    generic map(
        N_CTRL       => 3,
        N_STAT       => 0
    )
    port map(
        clk          => ipb_clk,
        reset        => ipb_rst,
        ipbus_in     => ipb_to_slaves(N_SLV_CONTROL),
        ipbus_out    => ipb_from_slaves(N_SLV_CONTROL),
        -- d         => link_aggregator_status_registers,
        q            => control_registers
    );

    --==============================--
    RateMonitoringInstance: entity work.ipbus_ctrlreg_v
    --==============================--
    generic map(
        N_CTRL       => 0,
        N_STAT       => 2
    )
    port map(
        clk          => ipb_clk,
        reset        => ipb_rst,
        ipbus_in     => ipb_to_slaves(N_SLV_RATE_MONITORING),
        ipbus_out    => ipb_from_slaves(N_SLV_RATE_MONITORING),
        d            => rate_monitoring_registers
    );

    --==============================--
    channel_select: entity work.ipbus_dc_fabric_sel
    --==============================--
    generic map(
        SEL_WIDTH       => 7
    )
    port map(
        clk             => ipb_clk,
        rst             => ipb_rst,
        sel             => channel_sel(0)(6 downto 0),
        ipb_in          => ipb_to_slaves(N_SLV_MONITORING),
        ipb_out         => ipb_from_slaves(N_SLV_MONITORING),
        ipbdc_out       => ipb_chain(0),
        ipbdc_in        => ipb_chain(8)
    );

    --==============================--
    pInputLinkMask: process(clk_p)
    --==============================--
    begin
        if rising_edge(clk_p) then
            for i in 0 to N_LINKS - 1 loop
                if control_registers(0)(i) = '1' then
                    masked_links(i) <= links_in(i);
                else
                    masked_links(i) <= LWORD_NULL;
                end if;
            end loop;
        end if;
    end process;

    --==============================--
    gMonitoringRegisters : for i in 0 to 7 generate
    --==============================--
        signal ipb_to_channel   : ipb_wbus;
        signal ipb_from_channel : ipb_rbus;
        signal din              : ipb_reg_v(6 - 1 downto 0) := (others => (others => '0'));
    begin
        --==============================--
        channel_node: entity work.ipbus_dc_node
        --==============================--
        generic map(
            I_SLV       => i,
            SEL_WIDTH   => 7,
            PIPELINE    => false
        )
        port map(
            clk         => ipb_clk,
            rst         => ipb_rst,
            ipb_out     => ipb_to_channel,
            ipb_in      => ipb_from_channel,
            ipbdc_in    => ipb_chain(i),
            ipbdc_out   => ipb_chain(i + 1)
        );

        --==============================--
        MProcessorMonitoringInstance: entity work.ipbus_ctrlreg_v
        --==============================--
        generic map(
            N_CTRL       => 0,
            N_STAT       => 6
        )
        port map(
            clk          => ipb_clk,
            reset        => ipb_rst,
            ipbus_in     => ipb_to_channel,
            ipbus_out    => ipb_from_channel,
            d            => din
        );

        din(N_LINKS - 1 downto 0) <= mprocessor_monitoring_registers((i+1)*N_LINKS - 1 downto i*N_LINKS);
    end generate gMonitoringRegisters;


    genThresholds: for i in 0 to N_LINKS - 1 generate
    begin
        thresholds_l(i) <= to_integer(unsigned(control_registers(1)(5*(i+1) - 1 downto 5*i)));
        thresholds_u(i) <= to_integer(unsigned(control_registers(2)(4*(i+1) - 1 downto 5*i)));
    end generate;


    --==============================--
    LinkAggregatorInstance: entity work.LinkAggregatorCore
    --==============================--
    generic map (
        N_LINKS => N_LINKS,
        INDEX   => INDEX
    )
    port map (
        --- Input Ports ---
        clk_p            => clk_p,
        reset            => control_registers(0)(6),
        packet_start     => packet_start,
        links_in         => masked_links,
        thresholds_l     => thresholds_l,
        thresholds_u     => thresholds_u,
        --- Output Ports ---
        links_out        => links_out,
        readout_reset    => readout_reset,
        --- Debug Ports ---
        -- status_bits      => link_aggregator_status_registers(3)(15 downto 0),
        non_empty_counts => mprocessor_monitoring_registers,
        bx_valid_count_out => bx_valid_count,
        debug_tracking_event_count => tracking_event_count,
        debug_tracking_stubs_out => tracking_stubs
    );
    
    rate_monitoring_registers(0) <= bx_valid_count(0);
    rate_monitoring_registers(1) <= tracking_event_count;

end Behavorial;
