library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.emp_data_types.all;

use work.ipbus.all;
use work.ipbus_reg_types.all;

entity LinkAggregator is
  generic (
    N_LINKS : integer
  );
  port (
    --- Input Ports ---
    clk_p              : in std_logic;
    reset              : in std_logic;
    packet_start       : in std_logic;
    links_in           : in ldata(N_LINKS - 1 downto 0);
    --- Output Ports ---
    links_out          : out ldata(7 downto 0) := (others => LWORD_NULL);
    uncompressed_stubs : out ldata(N_LINKS * 8 - 1 downto 0) := (others => LWORD_NULL);
    readout_reset      : out std_logic := '0';
    --- Debug Ports ---
    status_bits        : out std_logic_vector(15 downto 0);
    --- IPBus Ports ---
    clk                : in  std_logic;
    rst                : in  std_logic;
    ipb_in             : in  ipb_wbus;
    ipb_out            : out ipb_rbus;
    non_empty_counts   : out ipb_reg_v(8*N_LINKS - 1 downto 0)
  );
end LinkAggregator;

architecture uncompressed of LinkAggregator is

    constant cFIFOWidth : integer := 64;


    -- FIFO IP Cores

    COMPONENT link_agg_single_link_fifo
    PORT (
        --- Input Ports ---
        clk          : IN STD_LOGIC;
        srst         : IN STD_LOGIC;
        din          : IN STD_LOGIC_VECTOR(cFIFOWidth - 1 DOWNTO 0);
        wr_en        : IN STD_LOGIC;
        rd_en        : IN STD_LOGIC;
        --- Output Ports ---
        dout         : OUT STD_LOGIC_VECTOR(cFIFOWidth - 1 DOWNTO 0);
        full         : OUT STD_LOGIC;
        empty        : OUT STD_LOGIC;
        almost_empty : OUT STD_LOGIC;
        valid        : OUT STD_LOGIC;
        data_count   : OUT STD_LOGIC_VECTOR(3 DOWNTO 0);
        wr_rst_busy  : OUT STD_LOGIC;
        rd_rst_busy  : OUT STD_LOGIC
    );
    END COMPONENT;
    COMPONENT link_agg_single_link_fifo_fwft
    PORT (
        --- Input Ports ---
        clk          : IN STD_LOGIC;
        srst         : IN STD_LOGIC;
        din          : IN STD_LOGIC_VECTOR(cFIFOWidth - 1 DOWNTO 0);
        wr_en        : IN STD_LOGIC;
        rd_en        : IN STD_LOGIC;
        --- Output Ports ---
        dout         : OUT STD_LOGIC_VECTOR(cFIFOWidth - 1 DOWNTO 0);
        full         : OUT STD_LOGIC;
        empty        : OUT STD_LOGIC;
        almost_empty : OUT STD_LOGIC;
        valid        : OUT STD_LOGIC;
        data_count   : OUT STD_LOGIC_VECTOR(4 DOWNTO 0);
        wr_rst_busy  : OUT STD_LOGIC;
        rd_rst_busy  : OUT STD_LOGIC
    );
    END COMPONENT;
    COMPONENT link_agg_all_links_fifo
    PORT (
        --- Input Ports ---
        clk          : IN STD_LOGIC;
        srst         : IN STD_LOGIC;
        din          : IN STD_LOGIC_VECTOR(cFIFOWidth - 1 DOWNTO 0);
        wr_en        : IN STD_LOGIC;
        rd_en        : IN STD_LOGIC;
        --- Output Ports ---
        dout         : OUT STD_LOGIC_VECTOR(cFIFOWidth - 1 DOWNTO 0);
        full         : OUT STD_LOGIC;
        empty        : OUT STD_LOGIC;
        almost_empty : OUT STD_LOGIC;
        valid        : OUT STD_LOGIC;
        data_count   : OUT STD_LOGIC_VECTOR(5 DOWNTO 0);
        wr_rst_busy  : OUT STD_LOGIC;
        rd_rst_busy  : OUT STD_LOGIC
    );
    END COMPONENT;

    constant cBoxcarBx  : integer := 8;
    
    type tLinkDataArray is array(N_LINKS - 1 downto 0) of std_logic_vector(cFIFOWidth - 1 downto 0);
    type t2DDataArray is array(cBoxcarBx - 1 downto 0) of tLinkDataArray;
    type tEnableArray is array(cBoxcarBx - 1 downto 0) of std_logic_vector(N_LINKS - 1 downto 0);
    
    signal data              : t2DDataArray := (others => (others => (others => '0')));
    signal rd_en_route_array : tEnableArray := (others => (others => '0'));
    signal valid_route_array : tEnableArray := (others => (others => '0'));
    signal counter           : integer      := 0;

begin
    --==============================--
    genLinkCache : for i in N_LINKS - 1 downto 0 generate
    --==============================--
        signal wr_en_array_cache : std_logic_vector(cBoxcarBx - 1 downto 0) := (others => '0');
        signal data_in           : std_logic_vector(cFIFOWidth - 1 downto 0);       
    begin
    
        -- Send the data to the correct FIFO, using the Bx field of the stub
        pRouteInputData : process(clk_p)
        begin
            if rising_edge(clk_p) then
                wr_en_array_cache <= (others => '0');
                wr_en_array_cache(to_integer(unsigned(links_in(i).data(17 downto 15)))) <= links_in(i).valid;
                data_in <= std_logic_vector(to_unsigned(i, 4)) & links_in(i).data(cFIFOWidth - 1 - 4 downto 0);
            end if;
        end process;


        --==============================--
        genLinkBxCache : for j in cBoxcarBx - 1 downto 0 generate
        --==============================--
            signal dout_cache                                                                         : std_logic_vector(cFIFOWidth - 1 downto 0);
            signal rd_en_cache, full_cache, empty_cache, valid_cache, almost_empty_cache              : std_logic                                 := '0';
            signal count_cache                                                                        : std_logic_vector(3 downto 0);
            
            signal wr_en_route, rd_en_route, empty_route, full_route, valid_route, almost_empty_route : std_logic                                 := '0';
            signal din_route, dout_route                                                              : std_logic_vector(cFIFOWidth - 1 downto 0) := (others => '0');
            signal count_route                                                                        : std_logic_vector(4 downto 0);
            signal route_reset                                                                        : std_logic                                 := '0';

        begin
            
            pTransferToRoutingFIFO : process(clk_p)
            begin
                if rising_edge(clk_p) then
                    -- If FIFO is not empty at start of packet, enable readout to routing FIFO
                    if packet_start = '1' then
                        if empty_cache = '0' then
                            rd_en_cache <= '1';
                        end if;
                    end if;
                    if rd_en_cache = '1' then
                        -- Use almost empty flag to avoid underflow
                        if almost_empty_cache = '1' then
                            rd_en_cache <= '0';
                        end if;
                    end if;
                end if;
            end process;
            
            -- Enable writing of routing FIFO using the validflag of the cache FIFO
            wr_en_route <= valid_cache;
            din_route <= dout_cache;
            
            SingleLinkCacheInstance : link_agg_single_link_fifo
            PORT MAP (
                --- Input Ports ---
                clk          => clk_p,
                srst         => reset,
                din          => data_in,
                wr_en        => wr_en_array_cache(j),
                rd_en        => rd_en_cache,
                --- Output Ports ---
                dout         => dout_cache,
                full         => full_cache,
                empty        => empty_cache,
                valid        => valid_cache,
                almost_empty => almost_empty_cache,
                data_count   => count_cache
            );
            SingleLinkRoutingFIFOInstance : link_agg_single_link_fifo_fwft
            PORT MAP (
                --- Input Ports ---
                clk        => clk_p,
                srst       => route_reset, -- Reset FIFO for every packet
                din        => din_route,
                wr_en      => wr_en_route,
                rd_en      => rd_en_route_array(j)(i),
                --- Output Ports ---
                dout       => data(j)(i),
                full       => full_route,
                empty      => empty_route,
                valid      => valid_route_array(j)(i),
                data_count => count_route
            );

            route_reset <= reset or packet_start;
                
        end generate;
     end generate;
     
     genAggregatedFIFOs : for i in 7 downto 0 generate
        signal pointer : integer                       := 0;
        signal din     : std_logic_vector(cFIFOWidth - 1 downto 0) := (others => '0');
        
        signal wr_en, rd_en, full, empty, almost_empty, valid : std_logic                     := '0';
        signal dout                                           : std_logic_vector(cFIFOWidth - 1 downto 0) := (others => '0');
        signal data_count                                     : std_logic_vector(5 downto 0);
        signal all_links_reset                                : std_logic                     := '0';

     begin
        pIteratePointer : process(clk_p)
        begin
            if rising_edge(clk_p) then
                -- Loop pointer over all the links
                if pointer = N_LINKS - 1 then
                    pointer <= 0;
                else
                    pointer <= pointer + 1;
                end if;
                -- Readout value of FIFO which is currently pointed to
                rd_en_route_array(i) <= (others => '0');
                rd_en_route_array(i)(pointer) <= valid_route_array(i)(pointer);
                din <= data(i)(pointer);
                
                -- If counter is at the end of a packet, then start to readout the contents of the aggregated FIFOs
                if counter = 48 then
                    rd_en <= '1';
                end if;
                if rd_en = '1' then
                    if almost_empty = '1' then
                        rd_en <= '0';
                    end if;
                end if;

                if valid = '1' then
                    links_out(i).valid <= '1';
                    links_out(i).data <= dout;
                    links_out(i).strobe <= '1';
                else
                    links_out(i).valid <= '0';
                    links_out(i).data <= (others => '0');
                    links_out(i).strobe <= '1';
                end if;               
                wr_en <= valid_route_array(i)(pointer);
            end if;
            
        end process;
        
        
        AllLinksFIFOInstance : link_agg_all_links_fifo
        PORT MAP (
            --- Input Ports ---
            clk          => clk_p,
            srst         => all_links_reset, -- Reset FIFO on every packet
            din          => din,
            wr_en        => wr_en,
            rd_en        => rd_en,
            --- Output Ports ---
            dout         => dout,
            full         => full,
            empty        => empty,
            almost_empty => almost_empty,
            valid        => valid,
            data_count   => data_count
        );

        all_links_reset <= packet_start or reset;

     end generate;
     
     pReadoutReset : process(clk_p)
     begin
        if rising_edge(clk_p) then
            if packet_start = '1' then
                counter <= 0;
            else
                counter <= counter + 1;
            end if;
            -- If counter is at the end of a packet, then start to readout the contents of the aggregated FIFOs
            readout_reset <= '0';
            if counter = 64 then
                counter <= counter;
            elsif counter = 48 then
                readout_reset <= '1';
            end if;
        end if;
    end process pReadoutReset;

end uncompressed;
