library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.emp_data_types.all;

use work.ipbus.all;
use work.ipbus_reg_types.all;

use work.ipbus_decode_mprocessor.all;

use work.dtc_link_maps.all;
use work.mprocessor_constants.all;
use work.dtc_constants.all;
use work.dtc_data_types.all;


entity MProcessor is
    generic (
        N_LINKS : integer;
        INDEX   : integer := 0
    );
    port (
        --- Input Ports ---
        clk_p              : in  std_logic;
        links_in           : in  ldata(N_LINKS - 1 downto 0);
        header_start_array : in  tHeaderStartArray                                                := (others => (others => '0'));
        payload_header     : in  std_logic_vector(127 downto 0)                                   := (others => '0');
        gbe_backpressure   : in  std_logic                                                        := '0';
        srst               : in  std_logic                                                        := '0';
        --- Output Ports ---
        link_out           : out lword                                                            := LWORD_NULL;
        --- IPBus Ports ---
        ipb_clk            : in  std_logic;
        ipb_rst            : in  std_logic;
        ipb_in             : in  ipb_wbus;
        ipb_out            : out ipb_rbus;
        --- Debug Ports ---
        debug_readout_reset : out std_logic := '0'
    );
end MProcessor;


architecture Behavorial of MProcessor is

    signal ipb_to_slaves           : ipb_wbus_array(N_SLAVES - 1 downto 0);
    signal ipb_from_slaves         : ipb_rbus_array(N_SLAVES - 1 downto 0);

    signal packet_start            : std_logic                   := '0';
    signal readout_reset           : std_logic                   := '0';
    signal aggregated_stubs        : ldata(7 downto 0)           := (others => LWORD_NULL);
    signal link_aggregator_input   : ldata(N_LINKS - 1 downto 0) := (others => LWORD_NULL);

    signal status_registers        : ipb_reg_v(4 - 1 downto 0)   := (others => (others => '0'));
    signal control_registers       : ipb_reg_v(5 - 1 downto 0)   := (others => (others => '0'));

    signal super_id                : std_logic_vector(31 downto 0) := (others => '0');

    signal occupancy_trigger_window_lower             : std_logic_vector(31 downto 0)            := (others => '0');
    signal occupancy_trigger_window_upper             : std_logic_vector(3 downto 0)             := (others => '0');
    signal occupancy_trigger_window                   : std_logic_vector(36 - 1 downto 0)        := X"0ffffffff";
    signal occupancy_max_value                        : std_logic_vector(32 - 1 downto 0)        := (others => '0');
    signal occupancy_histogram_reset                  : std_logic                                := '0';
    signal debug_packet_size                          : lword                                    := LWORD_NULL;

    signal tracking_stubs : ldata(5 downto 0) := (others => LWORD_NULL);

begin

    --==============================--
    -- IPBus fabric
    --==============================--


    --==============================--
    fabric: entity work.ipbus_fabric_sel
    --==============================--
    generic map(
        NSLV            => N_SLAVES,
        SEL_WIDTH       => IPBUS_SEL_WIDTH
    )
    port map(
        ipb_in          => ipb_in,
        ipb_out         => ipb_out,
        sel             => ipbus_sel_mprocessor(ipb_in.ipb_addr),
        ipb_to_slaves   => ipb_to_slaves,
        ipb_from_slaves => ipb_from_slaves
    );

    --==============================--
    MProcessorControlInstance: entity work.ipbus_ctrlreg_v
    --==============================--
    generic map(
        N_CTRL    => 5,
        N_STAT    => 4
    )
    port map(
        clk       => ipb_clk,
        reset     => ipb_rst,
        ipbus_in  => ipb_to_slaves(N_SLV_CSR),
        ipbus_out => ipb_from_slaves(N_SLV_CSR),
        d         => status_registers,
        q         => control_registers
    );

    --==============================--
    LinkAggregatorInstance: entity work.LinkAggregatorIPBus
    --==============================--
    generic map (
        N_LINKS       => N_LINKS,
        INDEX         => INDEX
    )
    port map (
        --- Input Ports ---
        clk_p          => clk_p,
        packet_start   => packet_start,
        links_in       => link_aggregator_input,
        --- Output Ports ---
        links_out      => aggregated_stubs,
        readout_reset  => readout_reset,
        tracking_stubs => tracking_stubs,
        --- IPBus Ports ---
        ipb_clk        => ipb_clk,
        ipb_rst        => ipb_rst,
        ipb_in         => ipb_to_slaves(N_SLV_LINK_AGGREGATOR),
        ipb_out        => ipb_from_slaves(N_SLV_LINK_AGGREGATOR)
    );

    --==============================--
    LinkCombinerInstance : entity work.LinkCombinerIPBus
    --==============================--
    generic map (
        N_INPUT_LINKS => 8
    )
    port map (
        --- Input Ports ---
        clk_p         => clk_p,
        reset         => readout_reset,
        links_in      => aggregated_stubs,
        packet_start  => packet_start,
        payload_header => payload_header,
        lff           => gbe_backpressure,
        output_srst   => srst,
        --- Output Ports ---
        link_out      => link_out,
        --- IPBus Ports ---
        ipb_clk       => ipb_clk,
        ipb_rst       => ipb_rst,
        ipb_in        => ipb_to_slaves(N_SLV_LINK_COMBINER),
        ipb_out       => ipb_from_slaves(N_SLV_LINK_COMBINER),
        --- Debug Ports ---
        super_id      => super_id,
        debug_packet_size => debug_packet_size
    );

    -- --==============================--
    -- TrackReconstructorInstance: entity work.TrackReconstructor
    -- --==============================--
    -- port map(
    --     clk_p => clk_p,
    --     reset => readout_reset,
    --     stubs_in => aggregated_stubs(0)
    -- );

    --==============================--
    TrackFinderInstance : entity work.TrackFinder
    --==============================--
    port map (
        -- Input Ports --
        clk_p => clk_p,
        stubs_in => tracking_stubs,
        header_start => packet_start,
        --- IPBus Ports ---
        ipb_clk             => ipb_clk,
        ipb_rst             => ipb_rst,
        ipb_in              => ipb_to_slaves(N_SLV_TRACK_FINDER),
        ipb_out             => ipb_from_slaves(N_SLV_TRACK_FINDER)
    );

    link_aggregator_input(N_LINKS - 1 downto 0) <= links_in;

    status_registers(0)(0) <= gbe_backpressure;
    status_registers(1)    <= super_id;
    packet_start           <= header_start_array(to_integer(unsigned(control_registers(0)(2 downto 0))))(0);
    debug_readout_reset    <= readout_reset;


    --==============================--
    PacketSizeHistogramInstance : entity work.IPBusHistogram
    --==============================--
    generic map (
        input_width => 8,
        bin_width   => 32,
        data_offset => 0
    )
    port map (
        --- Input Ports ---
        clk_p           => clk_p,
        data_in         => debug_packet_size,
        histogram_reset => occupancy_histogram_reset,
        --- Output Ports ---
        max_bin_value   => occupancy_max_value,
        --- IPBus Ports ---
        clk             => ipb_clk,
        rst             => ipb_rst,
        ipb_in          => ipb_to_slaves(N_SLV_OCCUPANCY_MEM),
        ipb_out         => ipb_from_slaves(N_SLV_OCCUPANCY_MEM)
    );

    --==============================--
    OccupancyHistogramResetter : entity work.HistogramResetter
    --==============================--
    port map (
        --- Input Ports ---
        clk_p           => clk_p,
        trigger_window  => occupancy_trigger_window,
        --- Output Ports ---
        histogram_reset => occupancy_histogram_reset
    );

    occupancy_trigger_window_lower <= control_registers(1);
    occupancy_trigger_window_upper <= control_registers(2)(3 downto 0);
    occupancy_trigger_window       <= occupancy_trigger_window_upper & occupancy_trigger_window_lower;

    status_registers(2)(32 - 1 downto 0) <= occupancy_max_value;


end Behavorial;
