----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date: 09/09/2023 11:30:24 PM
-- Design Name: 
-- Module Name: TrackFinder - Behavioral
-- Project Name: 
-- Target Devices: 
-- Tool Versions: 
-- Description: 
-- 
-- Dependencies: 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use IEEE.STD_LOGIC_MISC.ALL;

use work.emp_data_types.all;
use work.tracking_constants.all;

use work.ipbus.all;
use work.ipbus_reg_types.all;
use work.ipbus_decode_track_finder.all;


entity TrackFinder is
    port (
        clk_p        : in STD_LOGIC;
        stubs_in     : in ldata(5 downto 0) := (others => LWORD_NULL);
        header_start : in std_logic;
        --- IPBus Ports ---
        ipb_clk            : in  std_logic;
        ipb_rst            : in  std_logic;
        ipb_in             : in  ipb_wbus;
        ipb_out            : out ipb_rbus
    );
end TrackFinder;

architecture Behavioral of TrackFinder is

    signal ipb_to_slaves           : ipb_wbus_array(N_SLAVES - 1 downto 0);
    signal ipb_from_slaves         : ipb_rbus_array(N_SLAVES - 1 downto 0);

    signal status_registers        : ipb_reg_v(4 - 1 downto 0)   := (others => (others => '0'));
    signal control_registers       : ipb_reg_v(2 - 1 downto 0)   := (others => (others => '0'));

    signal stub_array        : tDecodedStubArray := (others => cNullDecodedStub);
    signal valid_event_array : std_logic_vector(5 downto 0);
    signal valid_event       : std_logic := '0';

    signal counter       : integer := 0;
    signal track_delta_x : integer := 0;
    signal track_delta_y : integer := 0;

    signal localx_histogram_input : lword := LWORD_NULL;
    signal deltax_histogram_input : lword := LWORD_NULL;
    signal localy_histogram_input : lword := LWORD_NULL;
    signal deltay_histogram_input : lword := LWORD_NULL;
    
    signal localx_trigger_window_lower             : std_logic_vector(31 downto 0)            := (others => '0');
    signal localx_trigger_window_upper             : std_logic_vector(3 downto 0)             := (others => '0');
    signal localx_trigger_window                   : std_logic_vector(36 - 1 downto 0)        := X"0ffffffff";
    signal localx_max_value                        : std_logic_vector(32 - 1 downto 0)        := (others => '0');
    signal localx_histogram_reset                  : std_logic                                := '0';

    signal deltax_max_value                        : std_logic_vector(32 - 1 downto 0)        := (others => '0');
    signal localy_max_value                        : std_logic_vector(32 - 1 downto 0)        := (others => '0');
    signal deltay_max_value                        : std_logic_vector(32 - 1 downto 0)        := (others => '0');

begin

    --==============================--
    fabric: entity work.ipbus_fabric_sel
    --==============================--
    generic map(
        NSLV            => N_SLAVES,
        SEL_WIDTH       => IPBUS_SEL_WIDTH
    )
    port map(
        ipb_in          => ipb_in,
        ipb_out         => ipb_out,
        sel             => ipbus_sel_track_finder(ipb_in.ipb_addr),
        ipb_to_slaves   => ipb_to_slaves,
        ipb_from_slaves => ipb_from_slaves
    );

    --==============================--
    TrackFinderControlInstance: entity work.ipbus_ctrlreg_v
    --==============================--
    generic map(
        N_CTRL    => 2,
        N_STAT    => 4
    )
    port map(
        clk       => ipb_clk,
        reset     => ipb_rst,
        ipbus_in  => ipb_to_slaves(N_SLV_CSR),
        ipbus_out => ipb_from_slaves(N_SLV_CSR),
        d         => status_registers,
        q         => control_registers
    );

    --==============================--
    genValidEventArray: for i in 0 to 5 generate
    --==============================--
    begin
        valid_event_array(i) <= stubs_in(i).valid;
    end generate;

    --==============================--
    pMain : process(clk_p)
    --==============================--
    begin
        if rising_edge(clk_p) then
            if header_start = '1' then
                valid_event <= and_reduce(valid_event_array);
                for i in 0 to 5 loop
                    if stubs_in(i).valid = '1' then
                        stub_array(i) <= convertStubtoDecodedStub(convertSLVtoStub(stubs_in(i).data));
                    end if;
                end loop;
            end if;
            valid_event <= and_reduce(valid_event_array) and header_start;

            if valid_event = '1' then
                track_delta_x <= stub_array(4).localx - stub_array(0).localx + 1;
            end if;
            if counter = 1 then
                if (track_delta_x > -127) and (track_delta_x < 127) then
                    deltax_histogram_input.valid <= '1';
                    deltax_histogram_input.data(7 downto 0) <= std_logic_vector(to_unsigned(track_delta_x + 128, 8));
                else
                    deltax_histogram_input <= LWORD_NULL;
                end if;
            else
                deltax_histogram_input <= LWORD_NULL;
            end if;

            if valid_event = '1' then
                track_delta_y <= stub_array(5).localx - stub_array(1).localx + 1;
            end if;
            if counter = 1 then
                if (track_delta_y > -127) and (track_delta_y < 127) then
                    deltay_histogram_input.valid <= '1';
                    deltay_histogram_input.data(7 downto 0) <= std_logic_vector(to_unsigned(track_delta_y + 128, 8));
                else
                    deltay_histogram_input <= LWORD_NULL;
                end if;
            else
                deltax_histogram_input <= LWORD_NULL;
            end if;

            if valid_event = '1' then
                localx_histogram_input.valid <= '1';
                localx_histogram_input.data(10 downto 0) <= std_logic_vector(to_unsigned(stub_array(0).localx, 11));
                localy_histogram_input.valid <= '1';
                localy_histogram_input.data(10 downto 0) <= std_logic_vector(to_unsigned(stub_array(1).localx, 11));
            else
                localx_histogram_input <= LWORD_NULL;
                localy_histogram_input <= LWORD_NULL;
            end if;
        end if;
    end process pMain;

    --==============================--
    pCounter : process(clk_p)
    --==============================--
    begin
        if rising_edge(clk_p) then
            if header_start = '1' then
                counter <= 0;
            else
                if valid_event = '1' then
                    counter <= 1;
                else
                    if counter > 0 then
                        counter <= counter + 1;
                    end if;
                end if;
            end if;
        end if;
    end process pCounter;


    --==============================--
    LocalXHistogramInstance : entity work.IPBusHistogram
    --==============================--
    generic map (
        input_width => 8,
        bin_width   => 32,
        data_offset => 3
    )
    port map (
        --- Input Ports ---
        clk_p           => clk_p,
        data_in         => localx_histogram_input,
        histogram_reset => localx_histogram_reset,
        --- Output Ports ---
        max_bin_value   => localx_max_value,
        --- IPBus Ports ---
        clk             => ipb_clk,
        rst             => ipb_rst,
        ipb_in          => ipb_to_slaves(N_SLV_TRACKING_LOCALX_MEM),
        ipb_out         => ipb_from_slaves(N_SLV_TRACKING_LOCALX_MEM)
    );

    --==============================--
    LocalXHistogramResetter : entity work.HistogramResetter
    --==============================--
    port map (
        --- Input Ports ---
        clk_p           => clk_p,
        trigger_window  => localx_trigger_window,
        --- Output Ports ---
        histogram_reset => localx_histogram_reset
    );

    --==============================--
    DeltaXHistogramInstance : entity work.IPBusHistogram
    --==============================--
    generic map (
        input_width => 8,
        bin_width   => 32,
        data_offset => 0
    )
    port map (
        --- Input Ports ---
        clk_p           => clk_p,
        data_in         => deltax_histogram_input,
        histogram_reset => localx_histogram_reset,
        --- Output Ports ---
        max_bin_value   => deltax_max_value,
        --- IPBus Ports ---
        clk             => ipb_clk,
        rst             => ipb_rst,
        ipb_in          => ipb_to_slaves(N_SLV_TRACKING_DELTAX_MEM),
        ipb_out         => ipb_from_slaves(N_SLV_TRACKING_DELTAX_MEM)
    );

    --==============================--
    LocalYHistogramInstance : entity work.IPBusHistogram
    --==============================--
    generic map (
        input_width => 8,
        bin_width   => 32,
        data_offset => 3
    )
    port map (
        --- Input Ports ---
        clk_p           => clk_p,
        data_in         => localy_histogram_input,
        histogram_reset => localx_histogram_reset,
        --- Output Ports ---
        max_bin_value   => localy_max_value,
        --- IPBus Ports ---
        clk             => ipb_clk,
        rst             => ipb_rst,
        ipb_in          => ipb_to_slaves(N_SLV_TRACKING_LOCALY_MEM),
        ipb_out         => ipb_from_slaves(N_SLV_TRACKING_LOCALY_MEM)
    );

    --==============================--
    DeltaYHistogramInstance : entity work.IPBusHistogram
    --==============================--
    generic map (
        input_width => 8,
        bin_width   => 32,
        data_offset => 0
    )
    port map (
        --- Input Ports ---
        clk_p           => clk_p,
        data_in         => deltay_histogram_input,
        histogram_reset => localx_histogram_reset,
        --- Output Ports ---
        max_bin_value   => deltay_max_value,
        --- IPBus Ports ---
        clk             => ipb_clk,
        rst             => ipb_rst,
        ipb_in          => ipb_to_slaves(N_SLV_TRACKING_DELTAY_MEM),
        ipb_out         => ipb_from_slaves(N_SLV_TRACKING_DELTAY_MEM)
    );

    localx_trigger_window_lower <= control_registers(0);
    localx_trigger_window_upper <= control_registers(1)(3 downto 0);
    localx_trigger_window       <= localx_trigger_window_upper & localx_trigger_window_lower;

    status_registers(0)(32 - 1 downto 0) <= localx_max_value;
    status_registers(1)(32 - 1 downto 0) <= deltax_max_value;
    status_registers(2)(32 - 1 downto 0) <= localy_max_value;
    status_registers(3)(32 - 1 downto 0) <= deltay_max_value;

end Behavioral;
